#include <Rinternals.h>

static void pp(int pre) {
	int i=0;
	while (i++<pre) Rprintf(" ");
}

#define F_HEX_STR    1
#define F_PRINT_VEC  2

static char *typename(SEXP v) {
	switch (TYPEOF(v)) {
		case NILSXP: return "NILSXP";
		case SYMSXP: return "SYMSXP";
		case LISTSXP: return "LISTSXP";
		case CLOSXP: return "CLOSXP";
		case ENVSXP: return "ENVSXP";
		case PROMSXP: return "PROMSXP";
		case LANGSXP: return "LANGSXP";
		case SPECIALSXP: return "SPECIALSXP";
		case BUILTINSXP: return "BUILTINSXP";
		case CHARSXP: return "CHARSXP";
		case LGLSXP: return "LGLSXP";
		case INTSXP: return "INTSXP";
		case REALSXP: return "REALSXP";
		case CPLXSXP: return "CPLXSXP";
		case STRSXP: return "STRSXP";
		case DOTSXP: return "DOTSXP";
		case ANYSXP: return "ANYSXP";
		case VECSXP: return "VECSXP";
		case EXPRSXP: return "EXPRSXP";
		case BCODESXP: return "BCODESXP";
		case EXTPTRSXP: return "EXTPTRSXP";
		case WEAKREFSXP: return "WEAKREFSXP";
	        case RAWSXP: return "RAWSXP";
	        case S4SXP: return "S4SXP";
		default:
			return "<unknown>";
	}
}

static void inspect(int pre, SEXP v, int deep, int maxelts, unsigned int flags) {
	int a=0;
 	pp(pre); Rprintf("@%p %02d %s [", v, TYPEOF(v), typename(v));
	if (OBJECT(v)) { a=1; Rprintf("OBJ"); }
	if (MARK(v)) { if (a) Rprintf(","); Rprintf("MARK"); a=1; }
	if (NAMED(v)) { if (a) Rprintf(","); Rprintf("NAM(%d)",NAMED(v)); a=1; }
	if (LEVELS(v)) { if (a) Rprintf(","); Rprintf("GP(0x%x)",LEVELS(v)); a=1; }
	if (IS_S4_OBJECT(v)) { if (a) Rprintf(","); Rprintf("S4"); a=1; }
	if (ATTRIB(v) && ATTRIB(v)!=R_NilValue) { if (a) Rprintf(","); Rprintf("ATT"); a=1; }
	Rprintf("] ");
	switch (TYPEOF(v)) {
		case VECSXP: case STRSXP: case LGLSXP: case INTSXP:
		case REALSXP: case CPLXSXP: case EXPRSXP: case RAWSXP:
			Rprintf("(len=%d, tl=%d)", LENGTH(v), TRUELENGTH(v));
	}
	if (TYPEOF(v)==CHARSXP) {
		Rprintf("\"%s\"", CHAR(v));
		if (flags & F_HEX_STR) {
			Rprintf(" [");
			const unsigned char *c = (unsigned char*) CHAR(v);
			while (*c) Rprintf("%02x ", (int) *(c++));
			Rprintf("]");
		}
	}
	if (TYPEOF(v)==SYMSXP)
		Rprintf("\"%s\"", CHAR(PRINTNAME(v)));
	Rprintf("\n");
	if (deep) switch (TYPEOF(v)) {
		case VECSXP: case STRSXP: case EXPRSXP:
		{
			int i=0;
			while (i<LENGTH(v) && i<maxelts) {
			  inspect(pre+2, VECTOR_ELT(v, i), 1, maxelts, flags);
				i++;
			}
			if (i<LENGTH(v)) { pp(pre+2); Rprintf("...\n"); }
		}
			break;
		case LGLSXP: case INTSXP:
			if (flags & F_PRINT_VEC) {
				unsigned int i = 0, j = LENGTH(v);
				int *val = INTEGER(v);
				pp(pre + 2);
				while (i < j && i < maxelts) {
					Rprintf("%s%d", (i > 0) ? ", " : "", val[i]);
					i++;
				}
				if (i < j) Rprintf(", ...");
				Rprintf("\n");
			}
			break;
		case REALSXP:
			if (flags & F_PRINT_VEC) {
				unsigned int i = 0, j = LENGTH(v);
				double *val = REAL(v);
				pp(pre + 2);
				while (i < j && i < maxelts) {
					Rprintf("%s%g", (i > 0) ? ", " : "", val[i]);
					i++;
				}
				if (i < j) Rprintf(", ...");
				Rprintf("\n");
			}
			break;
		case CPLXSXP:
			break;
		case LISTSXP: case LANGSXP:
		{
			SEXP lc=v;
			while (lc!=R_NilValue) {
			  inspect(pre+2, CAR(lc), 1, maxelts, flags);
				if (TAG(lc) && TAG(lc)!=R_NilValue) {
					pp(pre+1); Rprintf("TAG:\n");
					inspect(pre+4, TAG(lc), 1, maxelts, flags);
				}
				lc=CDR(lc);
			}
		}
			break;
		case ENVSXP:
			pp(pre); Rprintf("FRAME:\n");
			inspect(pre+2, FRAME(v), 1, maxelts, flags);
			pp(pre); Rprintf("ENCLOS:\n");
			inspect(pre+2, ENCLOS(v), 0, maxelts, flags);
			pp(pre); Rprintf("HASHTAB:\n");
			inspect(pre+2, HASHTAB(v), 1, maxelts, flags);
			break;
			
		case CLOSXP:
			pp(pre); Rprintf("FORMALS:\n");
			inspect(pre+2, FORMALS(v), 1, maxelts, flags);
			pp(pre); Rprintf("BODY:\n");
			inspect(pre+2, BODY(v), 1, maxelts, flags);
			pp(pre); Rprintf("CLOENV:\n");
			inspect(pre+2, CLOENV(v), 0, maxelts, flags);
			break;
	}

	if (ATTRIB(v) && ATTRIB(v)!=R_NilValue) {
	  pp(pre); Rprintf("ATTRIB:\n"); inspect(pre+2, ATTRIB(v), 1, maxelts, flags);
	}
}

SEXP myInspect(SEXP val, SEXP sMax, SEXP sPrimElt, SEXP sHexStr) {
	int maxelts = asInteger(sMax);
	unsigned int flags = 0;
	if (TYPEOF(sPrimElt) == LGLSXP && LENGTH(sPrimElt) > 0 && LOGICAL(sPrimElt)[0] == 1)
		flags |= F_PRINT_VEC;
	if (TYPEOF(sHexStr) == LGLSXP && LENGTH(sHexStr) > 0 && LOGICAL(sHexStr)[0] == 1)
		flags |= F_HEX_STR;
	if (maxelts < 1) maxelts = 1000000;
	inspect(0, val, 1, maxelts, flags);
	return R_NilValue;
}
